@@ -22,19 +22,24 @@ import (
 	"errors"
 	"fmt"
 	"log"
+	"os"
+	"path/filepath"
 	goruntime "runtime"
 	"sort"
+	"strconv"
 	"strings"
 	"time"
 
-	"k8s.io/kubernetes/pkg/kubelet/types"
-
 	"github.com/docker/go-units"
 	godigest "github.com/opencontainers/go-digest"
 	"github.com/sirupsen/logrus"
 	"github.com/urfave/cli/v2"
+	errorUtils "k8s.io/apimachinery/pkg/util/errors"
 	internalapi "k8s.io/cri-api/pkg/apis"
 	pb "k8s.io/cri-api/pkg/apis/runtime/v1"
+	"k8s.io/kubelet/pkg/types"
+
+	"sigs.k8s.io/cri-tools/pkg/framework"
 )
 
 type containerByCreated []*pb.Container
@@ -82,6 +87,9 @@ type pullOptions struct {
 	// Username to use for accessing the registry
 	// password will be requested on the command line
 	username string
+
+	// timeout is the maximum time used for the image pull
+	timeout time.Duration
 }
 
 var createPullFlags = []cli.Flag{
@@ -95,19 +103,27 @@ var createPullFlags = []cli.Flag{
 	},
 	&cli.StringFlag{
 		Name:  "creds",
-		Value: "",
 		Usage: "Use `USERNAME[:PASSWORD]` for accessing the registry",
 	},
 	&cli.StringFlag{
 		Name:  "auth",
-		Value: "",
 		Usage: "Use `AUTH_STRING` for accessing the registry. AUTH_STRING is a base64 encoded 'USERNAME[:PASSWORD]'",
 	},
 	&cli.StringFlag{
 		Name:  "username",
-		Value: "",
 		Usage: "Use `USERNAME` for accessing the registry. The password will be requested on the command line",
 	},
+	&cli.DurationFlag{
+		Name:    "cancel-timeout",
+		Aliases: []string{"T"},
+		Usage:   "Seconds to wait for a container create request to complete before cancelling the request",
+	},
+	&cli.DurationFlag{
+		Name:    "pull-timeout",
+		Aliases: []string{"pt"},
+		Usage:   "Maximum time to be used for pulling the image, disabled if set to 0s",
+		EnvVars: []string{"CRICTL_PULL_TIMEOUT"},
+	},
 }
 
 var runPullFlags = []cli.Flag{
@@ -121,37 +137,84 @@ var runPullFlags = []cli.Flag{
 	},
 	&cli.StringFlag{
 		Name:  "creds",
-		Value: "",
 		Usage: "Use `USERNAME[:PASSWORD]` for accessing the registry",
 	},
 	&cli.StringFlag{
 		Name:  "auth",
-		Value: "",
 		Usage: "Use `AUTH_STRING` for accessing the registry. AUTH_STRING is a base64 encoded 'USERNAME[:PASSWORD]'",
 	},
 	&cli.StringFlag{
 		Name:  "username",
-		Value: "",
 		Usage: "Use `USERNAME` for accessing the registry. password will be requested",
 	},
+	&cli.StringFlag{
+		Name:    "runtime",
+		Aliases: []string{"r"},
+		Usage:   "Runtime handler to use. Available options are defined by the container runtime.",
+	},
+	&cli.DurationFlag{
+		Name:    "cancel-timeout",
+		Aliases: []string{"T"},
+		Usage:   "Seconds to wait for a container create request before cancelling the request",
+	},
+	&cli.DurationFlag{
+		Name:    "pull-timeout",
+		Aliases: []string{"pt"},
+		Usage:   "Maximum time to be used for pulling the image, disabled if set to 0s",
+		EnvVars: []string{"CRICTL_PULL_TIMEOUT"},
+	},
 }
 
-var createContainerCommand = &cli.Command{
-	Name:      "create",
-	Usage:     "Create a new container",
-	ArgsUsage: "POD container-config.[json|yaml] pod-config.[json|yaml]",
-	Flags: append(createPullFlags, &cli.DurationFlag{
-		Name:    "cancel-timeout",
-		Aliases: []string{"T"},
-		Usage:   "Seconds to wait for a container create request to complete before cancelling the request",
-	}),
+var subcommands = []*cli.Command{{
+	Name:    "jsonschema",
+	Aliases: []string{"js"},
+	Usage:   "Display the JSON schema for the pod or container config",
+	Flags: []cli.Flag{
+		&cli.BoolFlag{
+			Name:    "pod",
+			Aliases: []string{"p"},
+			Usage:   "Print the pod JSON schema, which will be generated from the PodSandboxConfig of the CRI API compiled with this version of crictl",
+		},
+		&cli.BoolFlag{
+			Name:    "container",
+			Aliases: []string{"c"},
+			Usage:   "Print the container JSON schema, which will be generated from the ContainerConfig of the CRI API compiled with this version of crictl",
+		},
+	},
+	UseShortOptionHandling: true,
+	Action: func(c *cli.Context) error {
+		if !c.IsSet("pod") && !c.IsSet("container") {
+			return cli.ShowSubcommandHelp(c)
+		}
 
+		if c.IsSet("pod") {
+			if err := printJSONSchema(&pb.PodSandboxConfig{}); err != nil {
+				return fmt.Errorf("print pod sandbox config JSON schema: %w", err)
+			}
+		}
+
+		if c.IsSet("container") {
+			if err := printJSONSchema(&pb.ContainerConfig{}); err != nil {
+				return fmt.Errorf("print container config JSON schema: %w", err)
+			}
+		}
+
+		return nil
+	},
+}}
+
+var createContainerCommand = &cli.Command{
+	Name:        "create",
+	Usage:       "Create a new container",
+	ArgsUsage:   "POD container-config.[json|yaml] pod-config.[json|yaml]",
+	Flags:       createPullFlags,
+	Subcommands: subcommands,
 	Action: func(c *cli.Context) (err error) {
 		if c.Args().Len() != 3 {
 			return cli.ShowSubcommandHelp(c)
 		}
-		if c.Bool("no-pull") == true && c.Bool("with-pull") == true {
-			return errors.New("confict: no-pull and with-pull are both set")
+		if c.Bool("no-pull") && c.Bool("with-pull") {
+			return errors.New("conflict: no-pull and with-pull are both set")
 		}
 
 		withPull := (!c.Bool("no-pull") && PullImageOnCreate) || c.Bool("with-pull")
@@ -174,6 +237,7 @@ var createContainerCommand = &cli.Command{
 					creds:    c.String("creds"),
 					auth:     c.String("auth"),
 					username: c.String("username"),
+					timeout:  c.Duration("pull-timeout"),
 				},
 				timeout: c.Duration("cancel-timeout"),
 			},
@@ -199,14 +263,14 @@ var startContainerCommand = &cli.Command{
 	ArgsUsage: "CONTAINER-ID [CONTAINER-ID...]",
 	Action: func(c *cli.Context) error {
 		if c.NArg() == 0 {
-			return fmt.Errorf("ID cannot be empty")
+			return errors.New("ID cannot be empty")
 		}
 		runtimeClient, err := getRuntimeService(c, 0)
 		if err != nil {
 			return err
 		}
 
-		for i := 0; i < c.NArg(); i++ {
+		for i := range c.NArg() {
 			containerID := c.Args().Get(i)
 			if err := StartContainer(runtimeClient, containerID); err != nil {
 				return fmt.Errorf("starting the container %q: %w", containerID, err)
@@ -256,14 +320,14 @@ var updateContainerCommand = &cli.Command{
 	},
 	Action: func(c *cli.Context) error {
 		if c.NArg() == 0 {
-			return fmt.Errorf("ID cannot be empty")
+			return errors.New("ID cannot be empty")
 		}
 		runtimeClient, err := getRuntimeService(c, 0)
 		if err != nil {
 			return err
 		}
 
-		options := updateOptions{
+		options := &updateOptions{
 			CPUCount:           c.Int64("cpu-count"),
 			CPUMaximum:         c.Int64("cpu-maximum"),
 			CPUPeriod:          c.Int64("cpu-period"),
@@ -274,7 +338,7 @@ var updateContainerCommand = &cli.Command{
 			MemoryLimitInBytes: c.Int64("memory"),
 		}
 
-		for i := 0; i < c.NArg(); i++ {
+		for i := range c.NArg() {
 			containerID := c.Args().Get(i)
 			if err := UpdateContainerResources(runtimeClient, containerID, options); err != nil {
 				return fmt.Errorf("updating container resources for %q: %w", containerID, err)
@@ -298,14 +362,14 @@ var stopContainerCommand = &cli.Command{
 	},
 	Action: func(c *cli.Context) error {
 		if c.NArg() == 0 {
-			return fmt.Errorf("ID cannot be empty")
+			return errors.New("ID cannot be empty")
 		}
 		runtimeClient, err := getRuntimeService(c, 0)
 		if err != nil {
 			return err
 		}
 
-		for i := 0; i < c.NArg(); i++ {
+		for i := range c.NArg() {
 			containerID := c.Args().Get(i)
 			if err := StopContainer(runtimeClient, containerID, c.Int64("timeout")); err != nil {
 				return fmt.Errorf("stopping the container %q: %w", containerID, err)
@@ -326,6 +390,11 @@ var removeContainerCommand = &cli.Command{
 			Aliases: []string{"f"},
 			Usage:   "Force removal of the container, disregarding if running",
 		},
+		&cli.BoolFlag{
+			Name:    "keep-logs",
+			Aliases: []string{"k"},
+			Usage:   "Preserve the container log file and its rotations",
+		},
 		&cli.BoolFlag{
 			Name:    "all",
 			Aliases: []string{"a"},
@@ -340,7 +409,9 @@ var removeContainerCommand = &cli.Command{
 
 		ids := ctx.Args().Slice()
 		if ctx.Bool("all") {
-			r, err := runtimeClient.ListContainers(context.TODO(), nil)
+			r, err := InterruptableRPC(nil, func(ctx context.Context) ([]*pb.Container, error) {
+				return runtimeClient.ListContainers(ctx, nil)
+			})
 			if err != nil {
 				return err
 			}
@@ -351,43 +422,58 @@ var removeContainerCommand = &cli.Command{
 		}
 
 		if len(ids) == 0 {
+			if ctx.Bool("all") {
+				logrus.Info("No containers to remove")
+				return nil
+			}
 			return cli.ShowSubcommandHelp(ctx)
 		}
 
-		errored := false
+		funcs := []func() error{}
 		for _, id := range ids {
-			resp, err := runtimeClient.ContainerStatus(context.TODO(), id, false)
-			if err != nil {
-				logrus.Error(err)
-				errored = true
-				continue
-			}
-			if resp.GetStatus().GetState() == pb.ContainerState_CONTAINER_RUNNING {
-				if ctx.Bool("force") {
-					if err := StopContainer(runtimeClient, id, 0); err != nil {
-						logrus.Errorf("stopping the container %q failed: %v", id, err)
-						errored = true
-						continue
+			funcs = append(funcs, func() error {
+				resp, err := InterruptableRPC(nil, func(ctx context.Context) (*pb.ContainerStatusResponse, error) {
+					return runtimeClient.ContainerStatus(ctx, id, false)
+				})
+				if err != nil {
+					return fmt.Errorf("getting container status %q: %w", id, err)
+				}
+				if resp.GetStatus().GetState() == pb.ContainerState_CONTAINER_RUNNING {
+					if ctx.Bool("force") {
+						if err := StopContainer(runtimeClient, id, 0); err != nil {
+							return fmt.Errorf("stopping the container %q: %w", id, err)
+						}
+					} else {
+						return fmt.Errorf("container %q is running, please stop it first", id)
 					}
-				} else {
-					logrus.Errorf("container %q is running, please stop it first", id)
-					errored = true
-					continue
 				}
-			}
 
-			err = RemoveContainer(runtimeClient, id)
-			if err != nil {
-				logrus.Errorf("removing container %q failed: %v", id, err)
-				errored = true
-				continue
-			}
+				err = RemoveContainer(runtimeClient, id)
+				if err != nil {
+					return fmt.Errorf("removing container %q: %w", id, err)
+				} else if !ctx.Bool("keep-logs") {
+					logPath := resp.GetStatus().GetLogPath()
+					if logPath != "" {
+						logRotations, err := filepath.Glob(logPath + ".*")
+						if err != nil {
+							logRotations = []string{}
+						}
+						logRotations = append(logRotations, logPath)
+
+						for _, logFile := range logRotations {
+							err = os.Remove(logFile)
+							if err != nil {
+								logrus.Errorf("removing log file %s for container %q failed: %v", logFile, id, err)
+							}
+						}
+					}
+				}
+
+				return nil
+			})
 		}
 
-		if errored {
-			return fmt.Errorf("unable to remove container(s)")
-		}
-		return nil
+		return errorUtils.AggregateGoroutines(funcs...)
 	},
 }
 
@@ -410,22 +496,90 @@ var containerStatusCommand = &cli.Command{
 			Name:  "template",
 			Usage: "The template string is only used when output is go-template; The Template format is golang template",
 		},
+		&cli.StringFlag{
+			Name:  "name",
+			Usage: "Filter by container name regular expression pattern",
+		},
+		&cli.StringFlag{
+			Name:  "namespace",
+			Usage: "Filter by pod namespace regular expression pattern",
+		},
+		&cli.StringFlag{
+			Name:    "pod",
+			Aliases: []string{"p"},
+			Usage:   "Filter by pod id",
+		},
+		&cli.StringFlag{
+			Name:  "image",
+			Usage: "Filter by container image",
+		},
+		&cli.StringFlag{
+			Name:    "state",
+			Aliases: []string{"s"},
+			Usage:   "Filter by container state",
+		},
+		&cli.StringSliceFlag{
+			Name:  "label",
+			Usage: "Filter by key=value label",
+		},
+		&cli.BoolFlag{
+			Name:    "latest",
+			Aliases: []string{"l"},
+			Usage:   "Show the most recently created container (includes all states)",
+		},
+		&cli.IntFlag{
+			Name:    "last",
+			Aliases: []string{"n"},
+			Usage:   "Show last n recently created containers (includes all states). Set 0 for unlimited.",
+		},
 	},
 	Action: func(c *cli.Context) error {
-		if c.NArg() == 0 {
-			return fmt.Errorf("ID cannot be empty")
-		}
 		runtimeClient, err := getRuntimeService(c, 0)
 		if err != nil {
 			return err
 		}
 
-		for i := 0; i < c.NArg(); i++ {
-			containerID := c.Args().Get(i)
-			if err := ContainerStatus(runtimeClient, containerID, c.String("output"), c.String("template"), c.Bool("quiet")); err != nil {
-				return fmt.Errorf("getting the status of the container %q: %w", containerID, err)
+		ids := c.Args().Slice()
+
+		if len(ids) == 0 {
+			opts := &listOptions{
+				nameRegexp:         c.String("name"),
+				podID:              c.String("pod"),
+				podNamespaceRegexp: c.String("namespace"),
+				image:              c.String("image"),
+				state:              c.String("state"),
+				latest:             c.Bool("latest"),
+				last:               c.Int("last"),
+			}
+			opts.labels, err = parseLabelStringSlice(c.StringSlice("label"))
+			if err != nil {
+				return err
+			}
+
+			ctrs, err := ListContainers(runtimeClient, opts)
+			if err != nil {
+				return fmt.Errorf("listing containers: %w", err)
+			}
+			for _, ctr := range ctrs {
+				ids = append(ids, ctr.GetId())
 			}
 		}
+
+		if len(ids) == 0 {
+			logrus.Error("No IDs provided or nothing found per filter")
+			return cli.ShowSubcommandHelp(c)
+		}
+
+		if err := containerStatus(
+			runtimeClient,
+			ids,
+			c.String("output"),
+			c.String("template"),
+			c.Bool("quiet"),
+		); err != nil {
+			return fmt.Errorf("get the status of containers: %w", err)
+		}
+
 		return nil
 	},
 }
@@ -442,29 +596,24 @@ var listContainersCommand = &cli.Command{
 		},
 		&cli.StringFlag{
 			Name:  "id",
-			Value: "",
 			Usage: "Filter by container id",
 		},
 		&cli.StringFlag{
 			Name:  "name",
-			Value: "",
-			Usage: "filter by container name regular expression pattern",
+			Usage: "Filter by container name regular expression pattern",
 		},
 		&cli.StringFlag{
 			Name:    "pod",
 			Aliases: []string{"p"},
-			Value:   "",
 			Usage:   "Filter by pod id",
 		},
 		&cli.StringFlag{
 			Name:  "image",
-			Value: "",
 			Usage: "Filter by container image",
 		},
 		&cli.StringFlag{
 			Name:    "state",
 			Aliases: []string{"s"},
-			Value:   "",
 			Usage:   "Filter by container state",
 		},
 		&cli.StringSliceFlag{
@@ -480,7 +629,7 @@ var listContainersCommand = &cli.Command{
 			Name:    "output",
 			Aliases: []string{"o"},
 			Usage:   "Output format, One of: json|yaml|table",
-			Value:   "table",
+			Value:   outputTypeTable,
 		},
 		&cli.BoolFlag{
 			Name:    "all",
@@ -506,6 +655,10 @@ var listContainersCommand = &cli.Command{
 			Aliases: []string{"r"},
 			Usage:   "Show image path instead of image id",
 		},
+		&cli.StringFlag{
+			Name:  "namespace",
+			Usage: "Filter by pod namespace regular expression pattern",
+		},
 	},
 	Action: func(c *cli.Context) error {
 		if c.NArg() != 0 {
@@ -522,27 +675,28 @@ var listContainersCommand = &cli.Command{
 			return err
 		}
 
-		opts := listOptions{
-			id:               c.String("id"),
-			podID:            c.String("pod"),
-			state:            c.String("state"),
-			verbose:          c.Bool("verbose"),
-			quiet:            c.Bool("quiet"),
-			output:           c.String("output"),
-			all:              c.Bool("all"),
-			nameRegexp:       c.String("name"),
-			latest:           c.Bool("latest"),
-			last:             c.Int("last"),
-			noTrunc:          c.Bool("no-trunc"),
-			image:            c.String("image"),
-			resolveImagePath: c.Bool("resolve-image-path"),
+		opts := &listOptions{
+			id:                 c.String("id"),
+			podID:              c.String("pod"),
+			podNamespaceRegexp: c.String("namespace"),
+			state:              c.String("state"),
+			verbose:            c.Bool("verbose"),
+			quiet:              c.Bool("quiet"),
+			output:             c.String("output"),
+			all:                c.Bool("all"),
+			nameRegexp:         c.String("name"),
+			latest:             c.Bool("latest"),
+			last:               c.Int("last"),
+			noTrunc:            c.Bool("no-trunc"),
+			image:              c.String("image"),
+			resolveImagePath:   c.Bool("resolve-image-path"),
 		}
 		opts.labels, err = parseLabelStringSlice(c.StringSlice("label"))
 		if err != nil {
 			return err
 		}
 
-		if err = ListContainers(runtimeClient, imageClient, opts); err != nil {
+		if err = OutputContainers(runtimeClient, imageClient, opts); err != nil {
 			return fmt.Errorf("listing containers: %w", err)
 		}
 		return nil
@@ -550,25 +704,17 @@ var listContainersCommand = &cli.Command{
 }
 
 var runContainerCommand = &cli.Command{
-	Name:      "run",
-	Usage:     "Run a new container inside a sandbox",
-	ArgsUsage: "container-config.[json|yaml] pod-config.[json|yaml]",
-	Flags: append(runPullFlags, &cli.StringFlag{
-		Name:    "runtime",
-		Aliases: []string{"r"},
-		Usage:   "Runtime handler to use. Available options are defined by the container runtime.",
-	}, &cli.DurationFlag{
-		Name:    "timeout",
-		Aliases: []string{"t"},
-		Usage:   "Seconds to wait for a container create request before cancelling the request",
-	}),
-
+	Name:        "run",
+	Usage:       "Run a new container inside a sandbox",
+	ArgsUsage:   "container-config.[json|yaml] pod-config.[json|yaml]",
+	Flags:       runPullFlags,
+	Subcommands: subcommands,
 	Action: func(c *cli.Context) (err error) {
 		if c.Args().Len() != 2 {
 			return cli.ShowSubcommandHelp(c)
 		}
-		if c.Bool("no-pull") == true && c.Bool("with-pull") == true {
-			return errors.New("confict: no-pull and with-pull are both set")
+		if c.Bool("no-pull") && c.Bool("with-pull") {
+			return errors.New("conflict: no-pull and with-pull are both set")
 		}
 
 		withPull := (!DisablePullOnRun && !c.Bool("no-pull")) || c.Bool("with-pull")
@@ -589,8 +735,9 @@ var runContainerCommand = &cli.Command{
 				creds:    c.String("creds"),
 				auth:     c.String("auth"),
 				username: c.String("username"),
+				timeout:  c.Duration("pull-timeout"),
 			},
-			timeout: c.Duration("timeout"),
+			timeout: c.Duration("cancel-timeout"),
 		}
 
 		runtimeClient, err := getRuntimeService(c, opts.timeout)
@@ -614,19 +761,25 @@ var checkpointContainerCommand = &cli.Command{
 		&cli.StringFlag{
 			Name:    "export",
 			Aliases: []string{"e"},
-			Usage:   "Specify the name of the archive used to export the checkpoint image.",
+			Usage:   "Specify the name of the tar archive (/path/to/checkpoint.tar) used to export the checkpoint image.",
 		},
 	},
 	Action: func(c *cli.Context) error {
 		if c.NArg() == 0 {
-			return fmt.Errorf("ID cannot be empty")
+			return errors.New("ID cannot be empty")
+		}
+		if c.String("export") == "" {
+			return errors.New(
+				"cannot checkpoint a container without specifying the checkpoint destination. " +
+					"Use --export=/path/to/checkpoint.tar",
+			)
 		}
 		runtimeClient, err := getRuntimeService(c, 0)
 		if err != nil {
 			return err
 		}
 
-		for i := 0; i < c.NArg(); i++ {
+		for i := range c.NArg() {
 			containerID := c.Args().Get(i)
 			err := CheckpointContainer(
 				runtimeClient,
@@ -641,7 +794,7 @@ var checkpointContainerCommand = &cli.Command{
 	},
 }
 
-// RunContainer starts a container in the provided sandbox
+// RunContainer starts a container in the provided sandbox.
 func RunContainer(
 	iClient internalapi.ImageManagerService,
 	rClient internalapi.RuntimeService,
@@ -682,7 +835,6 @@ func CreateContainer(
 	rClient internalapi.RuntimeService,
 	opts createOptions,
 ) (string, error) {
-
 	config, err := loadContainerConfig(opts.configPath)
 	if err != nil {
 		return "", err
@@ -695,6 +847,11 @@ func CreateContainer(
 		}
 	}
 
+	image := config.GetImage().GetImage()
+	if config.Image.UserSpecifiedImage == "" {
+		config.Image.UserSpecifiedImage = image
+	}
+
 	// When there is a with-pull request or the image default mode is to
 	// pull-image-on-create(true) and no-pull was not set we pull the image when
 	// they ask for a create as a helper on the cli to reduce extra steps. As a
@@ -707,10 +864,21 @@ func CreateContainer(
 		}
 
 		// Try to pull the image before container creation
-		image := config.GetImage().GetImage()
-		ann := config.GetImage().GetAnnotations()
-		if _, err := PullImageWithSandbox(iClient, image, auth, podConfig, ann); err != nil {
-			return "", err
+		images := []string{image}
+		logrus.Infof("Pulling container image: %s", image)
+
+		// Add possible OCI volume mounts
+		for _, m := range config.Mounts {
+			if m.Image != nil && m.Image.Image != "" {
+				logrus.Infof("Pulling image %s to be mounted to container path: %s", image, m.ContainerPath)
+				images = append(images, m.Image.Image)
+			}
+		}
+
+		for _, image := range images {
+			if _, err := PullImageWithSandbox(iClient, image, auth, podConfig, config.GetImage().GetAnnotations(), opts.pullOptions.timeout); err != nil {
+				return "", err
+			}
 		}
 	}
 
@@ -720,7 +888,9 @@ func CreateContainer(
 		SandboxConfig: podConfig,
 	}
 	logrus.Debugf("CreateContainerRequest: %v", request)
-	r, err := rClient.CreateContainer(context.TODO(), opts.podID, config, podConfig)
+	r, err := InterruptableRPC(nil, func(ctx context.Context) (string, error) {
+		return rClient.CreateContainer(ctx, opts.podID, config, podConfig)
+	})
 	logrus.Debugf("CreateContainerResponse: %v", r)
 	if err != nil {
 		return "", err
@@ -732,9 +902,11 @@ func CreateContainer(
 // the returned StartContainerResponse.
 func StartContainer(client internalapi.RuntimeService, id string) error {
 	if id == "" {
-		return fmt.Errorf("ID cannot be empty")
+		return errors.New("ID cannot be empty")
 	}
-	if err := client.StartContainer(context.TODO(), id); err != nil {
+	if _, err := InterruptableRPC(nil, func(ctx context.Context) (any, error) {
+		return nil, client.StartContainer(ctx, id)
+	}); err != nil {
 		return err
 	}
 	fmt.Println(id)
@@ -764,14 +936,14 @@ type updateOptions struct {
 
 // UpdateContainerResources sends an UpdateContainerResourcesRequest to the server, and parses
 // the returned UpdateContainerResourcesResponse.
-func UpdateContainerResources(client internalapi.RuntimeService, id string, opts updateOptions) error {
+func UpdateContainerResources(client internalapi.RuntimeService, id string, opts *updateOptions) error {
 	if id == "" {
-		return fmt.Errorf("ID cannot be empty")
+		return errors.New("ID cannot be empty")
 	}
 	request := &pb.UpdateContainerResourcesRequest{
 		ContainerId: id,
 	}
-	if goruntime.GOOS != "windows" {
+	if goruntime.GOOS != framework.OSWindows {
 		request.Linux = &pb.LinuxContainerResources{
 			CpuPeriod:          opts.CPUPeriod,
 			CpuQuota:           opts.CPUQuota,
@@ -791,7 +963,9 @@ func UpdateContainerResources(client internalapi.RuntimeService, id string, opts
 	}
 	logrus.Debugf("UpdateContainerResourcesRequest: %v", request)
 	resources := &pb.ContainerResources{Linux: request.Linux}
-	if err := client.UpdateContainerResources(context.TODO(), id, resources); err != nil {
+	if _, err := InterruptableRPC(nil, func(ctx context.Context) (any, error) {
+		return nil, client.UpdateContainerResources(ctx, id, resources)
+	}); err != nil {
 		return err
 	}
 	fmt.Println(id)
@@ -802,34 +976,39 @@ func UpdateContainerResources(client internalapi.RuntimeService, id string, opts
 // the returned StopContainerResponse.
 func StopContainer(client internalapi.RuntimeService, id string, timeout int64) error {
 	if id == "" {
-		return fmt.Errorf("ID cannot be empty")
+		return errors.New("ID cannot be empty")
 	}
-	if err := client.StopContainer(context.TODO(), id, timeout); err != nil {
+	logrus.Debugf("Stopping container: %s (timeout = %v)", id, timeout)
+	if _, err := InterruptableRPC(nil, func(ctx context.Context) (any, error) {
+		return nil, client.StopContainer(ctx, id, timeout)
+	}); err != nil {
 		return err
 	}
 	fmt.Println(id)
 	return nil
 }
 
-// CheckpointContainer sends a CheckpointContainerRequest to the server
+// CheckpointContainer sends a CheckpointContainerRequest to the server.
 func CheckpointContainer(
 	rClient internalapi.RuntimeService,
-	ID string,
+	id string,
 	export string,
 ) error {
-	if ID == "" {
+	if id == "" {
 		return errors.New("ID cannot be empty")
 	}
 	request := &pb.CheckpointContainerRequest{
-		ContainerId: ID,
+		ContainerId: id,
 		Location:    export,
 	}
 	logrus.Debugf("CheckpointContainerRequest: %v", request)
-	err := rClient.CheckpointContainer(context.TODO(), request)
+	_, err := InterruptableRPC(nil, func(ctx context.Context) (*pb.ImageFsInfoResponse, error) {
+		return nil, rClient.CheckpointContainer(ctx, request)
+	})
 	if err != nil {
 		return err
 	}
-	fmt.Println(ID)
+	fmt.Println(id)
 	return nil
 }
 
@@ -837,9 +1016,12 @@ func CheckpointContainer(
 // the returned RemoveContainerResponse.
 func RemoveContainer(client internalapi.RuntimeService, id string) error {
 	if id == "" {
-		return fmt.Errorf("ID cannot be empty")
+		return errors.New("ID cannot be empty")
 	}
-	if err := client.RemoveContainer(context.TODO(), id); err != nil {
+	logrus.Debugf("Removing container: %s", id)
+	if _, err := InterruptableRPC(nil, func(ctx context.Context) (any, error) {
+		return nil, client.RemoveContainer(ctx, id)
+	}); err != nil {
 		return err
 	}
 	fmt.Println(id)
@@ -877,41 +1059,50 @@ func marshalContainerStatus(cs *pb.ContainerStatus) (string, error) {
 	return marshalMapInOrder(jsonMap, *cs)
 }
 
-// ContainerStatus sends a ContainerStatusRequest to the server, and parses
+// containerStatus sends a ContainerStatusRequest to the server, and parses
 // the returned ContainerStatusResponse.
-func ContainerStatus(client internalapi.RuntimeService, id, output string, tmplStr string, quiet bool) error {
+//
+//nolint:dupl // pods and containers are similar, but still different
+func containerStatus(client internalapi.RuntimeService, ids []string, output, tmplStr string, quiet bool) error {
 	verbose := !(quiet)
 	if output == "" { // default to json output
-		output = "json"
+		output = outputTypeJSON
 	}
-	if id == "" {
-		return fmt.Errorf("ID cannot be empty")
-	}
-	request := &pb.ContainerStatusRequest{
-		ContainerId: id,
-		Verbose:     verbose,
-	}
-	logrus.Debugf("ContainerStatusRequest: %v", request)
-	r, err := client.ContainerStatus(context.TODO(), id, verbose)
-	logrus.Debugf("ContainerStatusResponse: %v", r)
-	if err != nil {
-		return err
+	if len(ids) == 0 {
+		return errors.New("ID cannot be empty")
 	}
 
-	status, err := marshalContainerStatus(r.Status)
-	if err != nil {
-		return err
+	statuses := []statusData{}
+	for _, id := range ids {
+		request := &pb.ContainerStatusRequest{
+			ContainerId: id,
+			Verbose:     verbose,
+		}
+		logrus.Debugf("ContainerStatusRequest: %v", request)
+		r, err := InterruptableRPC(nil, func(ctx context.Context) (*pb.ContainerStatusResponse, error) {
+			return client.ContainerStatus(ctx, id, verbose)
+		})
+		logrus.Debugf("ContainerStatusResponse: %v", r)
+		if err != nil {
+			return fmt.Errorf("get container status: %w", err)
+		}
+
+		statusJSON, err := marshalContainerStatus(r.Status)
+		if err != nil {
+			return fmt.Errorf("marshal container status: %w", err)
+		}
+
+		if output == outputTypeTable {
+			outputContainerStatusTable(r, verbose)
+		} else {
+			statuses = append(statuses, statusData{json: statusJSON, info: r.Info})
+		}
 	}
 
-	switch output {
-	case "json", "yaml", "go-template":
-		return outputStatusInfo(status, r.Info, output, tmplStr)
-	case "table": // table output is after this switch block
-	default:
-		return fmt.Errorf("output option cannot be %s", output)
-	}
+	return outputStatusData(statuses, output, tmplStr)
+}
 
-	// output in table format
+func outputContainerStatusTable(r *pb.ContainerStatusResponse, verbose bool) {
 	fmt.Printf("ID: %s\n", r.Status.Id)
 	if r.Status.Metadata != nil {
 		if r.Status.Metadata.Name != "" {
@@ -950,13 +1141,11 @@ func ContainerStatus(client internalapi.RuntimeService, id, output string, tmplS
 	if verbose {
 		fmt.Printf("Info: %v\n", r.GetInfo())
 	}
-
-	return nil
 }
 
 // ListContainers sends a ListContainerRequest to the server, and parses
 // the returned ListContainerResponse.
-func ListContainers(runtimeClient internalapi.RuntimeService, imageClient internalapi.ImageManagerService, opts listOptions) error {
+func ListContainers(runtimeClient internalapi.RuntimeService, opts *listOptions) ([]*pb.Container, error) {
 	filter := &pb.ContainerFilter{}
 	if opts.id != "" {
 		filter.Id = opts.id
@@ -995,27 +1184,38 @@ func ListContainers(runtimeClient internalapi.RuntimeService, imageClient intern
 	if opts.labels != nil {
 		filter.LabelSelector = opts.labels
 	}
-	r, err := runtimeClient.ListContainers(context.TODO(), filter)
+	r, err := InterruptableRPC(nil, func(ctx context.Context) ([]*pb.Container, error) {
+		return runtimeClient.ListContainers(ctx, filter)
+	})
 	logrus.Debugf("ListContainerResponse: %v", r)
 	if err != nil {
-		return err
+		return nil, fmt.Errorf("call list containers RPC: %w", err)
+	}
+	return getContainersList(r, opts), nil
+}
+
+// OutputContainers sends a ListContainerRequest to the server, and parses
+// the returned ListContainerResponse for output.
+func OutputContainers(runtimeClient internalapi.RuntimeService, imageClient internalapi.ImageManagerService, opts *listOptions) error {
+	r, err := ListContainers(runtimeClient, opts)
+	if err != nil {
+		return fmt.Errorf("list containers: %w", err)
 	}
-	r = getContainersList(r, opts)
 
 	switch opts.output {
-	case "json":
+	case outputTypeJSON:
 		return outputProtobufObjAsJSON(&pb.ListContainersResponse{Containers: r})
-	case "yaml":
+	case outputTypeYAML:
 		return outputProtobufObjAsYAML(&pb.ListContainersResponse{Containers: r})
-	case "table":
+	case outputTypeTable:
 	// continue; output will be generated after the switch block ends.
 	default:
 		return fmt.Errorf("unsupported output format %q", opts.output)
 	}
 
-	display := newTableDisplay(20, 1, 3, ' ', 0)
+	display := newDefaultTableDisplay()
 	if !opts.verbose && !opts.quiet {
-		display.AddRow([]string{columnContainer, columnImage, columnCreated, columnState, columnName, columnAttempt, columnPodID, columnPodname})
+		display.AddRow([]string{columnContainer, columnImage, columnCreated, columnState, columnName, columnAttempt, columnPodID, columnPodName, columnNamespace})
 	}
 	for _, c := range r {
 		if match, err := matchesImage(imageClient, opts.image, c.GetImage().GetImage()); err != nil {
@@ -1030,10 +1230,14 @@ func ListContainers(runtimeClient internalapi.RuntimeService, imageClient intern
 
 		createdAt := time.Unix(0, c.CreatedAt)
 		ctm := units.HumanDuration(time.Now().UTC().Sub(createdAt)) + " ago"
+		podNamespace := getPodNamespaceFromLabels(c.Labels)
 		if !opts.verbose {
 			id := c.Id
-			image := c.Image.Image
 			podID := c.PodSandboxId
+			var image string
+			if c.Image != nil {
+				image = c.Image.Image
+			}
 			if !opts.noTrunc {
 				id = getTruncatedID(id, "")
 				podID = getTruncatedID(podID, "")
@@ -1045,18 +1249,21 @@ func ListContainers(runtimeClient internalapi.RuntimeService, imageClient intern
 			if opts.resolveImagePath {
 				orig, err := getRepoImage(imageClient, image)
 				if err != nil {
-					return fmt.Errorf("failed to fetch repo image %v", err)
+					return fmt.Errorf("failed to fetch repo image %w", err)
 				}
 				image = orig
 			}
 			podName := getPodNameFromLabels(c.Labels)
-			display.AddRow([]string{id, image, ctm, convertContainerState(c.State), c.Metadata.Name,
-				fmt.Sprintf("%d", c.Metadata.Attempt), podID, podName})
+			display.AddRow([]string{
+				id, image, ctm, convertContainerState(c.State), c.Metadata.Name,
+				strconv.FormatUint(uint64(c.Metadata.Attempt), 10), podID, podName, podNamespace,
+			})
 			continue
 		}
 
 		fmt.Printf("ID: %s\n", c.Id)
 		fmt.Printf("PodID: %s\n", c.PodSandboxId)
+		fmt.Printf("Namespace: %s\n", podNamespace)
 		if c.Metadata != nil {
 			if c.Metadata.Name != "" {
 				fmt.Printf("Name: %s\n", c.Metadata.Name)
@@ -1103,19 +1310,30 @@ func convertContainerState(state pb.ContainerState) string {
 	}
 }
 
-func getPodNameFromLabels(label map[string]string) string {
-	podName, ok := label[types.KubernetesPodNameLabel]
+func getPodNameFromLabels(labels map[string]string) string {
+	return getFromLabels(labels, types.KubernetesPodNameLabel)
+}
+
+func getPodNamespaceFromLabels(labels map[string]string) string {
+	return getFromLabels(labels, types.KubernetesPodNamespaceLabel)
+}
+
+func getFromLabels(labels map[string]string, label string) string {
+	value, ok := labels[label]
 	if ok {
-		return podName
+		return value
 	}
 	return "unknown"
 }
 
-func getContainersList(containersList []*pb.Container, opts listOptions) []*pb.Container {
+func getContainersList(containersList []*pb.Container, opts *listOptions) []*pb.Container {
 	filtered := []*pb.Container{}
 	for _, c := range containersList {
+		podNamespace := getPodNamespaceFromLabels(c.Labels)
 		// Filter by pod name/namespace regular expressions.
-		if matchesRegex(opts.nameRegexp, c.Metadata.Name) {
+		if c.Metadata != nil &&
+			matchesRegex(opts.nameRegexp, c.Metadata.Name) &&
+			matchesRegex(opts.podNamespaceRegexp, podNamespace) {
 			filtered = append(filtered, c)
 		}
 	}
