@@ -16,12 +16,23 @@
 
 set -euo pipefail
 
-go install k8s.io/release/cmd/release-notes@latest
+BUILD_DIR=build
+BINARY=$BUILD_DIR/release-notes
+VERSION=v0.17.8
 
-release-notes \
-    --discover minor-to-minor \
+mkdir -p $BUILD_DIR
+curl -sSfL --retry 5 --retry-delay 10 -o $BINARY \
+    https://storage.googleapis.com/k8s-artifacts-sig-release/kubernetes/release/$VERSION/release-notes-amd64-linux
+chmod +x $BINARY
+
+PREVIOUS_TAG=$(git tag | tail -2 | head -1)
+LATEST_TAG=$(git tag | tail -1)
+
+$BINARY \
     --org kubernetes-sigs \
     --repo cri-tools \
     --required-author "" \
-    --dependencies=false \
+    --branch master \
+    --start-rev "$PREVIOUS_TAG" \
+    --end-rev "$LATEST_TAG" \
     --output release-notes.md
