@@ -19,22 +19,19 @@ package validate
 import (
 	"bufio"
 	"context"
-	"encoding/json"
 	"errors"
-	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
+	"regexp"
 	"strings"
 	"time"
 
-	"github.com/docker/docker/daemon/logger/jsonfilelog/jsonlog"
-	"github.com/kubernetes-sigs/cri-tools/pkg/framework"
+	. "github.com/onsi/ginkgo/v2"
+	. "github.com/onsi/gomega"
 	internalapi "k8s.io/cri-api/pkg/apis"
 	runtimeapi "k8s.io/cri-api/pkg/apis/runtime/v1"
 
-	. "github.com/onsi/ginkgo/v2"
-	. "github.com/onsi/gomega"
+	"sigs.k8s.io/cri-tools/pkg/framework"
 )
 
 // streamType is the type of the stream.
@@ -76,9 +73,9 @@ var _ = framework.KubeDescribe("Container", func() {
 
 		AfterEach(func() {
 			By("stop PodSandbox")
-			rc.StopPodSandbox(context.TODO(), podID)
+			Expect(rc.StopPodSandbox(context.TODO(), podID)).NotTo(HaveOccurred())
 			By("delete PodSandbox")
-			rc.RemovePodSandbox(context.TODO(), podID)
+			Expect(rc.RemovePodSandbox(context.TODO(), podID)).NotTo(HaveOccurred())
 		})
 
 		It("runtime should support creating container [Conformance]", func() {
@@ -241,6 +238,22 @@ var _ = framework.KubeDescribe("Container", func() {
 			Expect(statFound(stats, secondContainerID)).To(BeTrue(), "Stats should be found")
 			Expect(statFound(stats, thirdContainerID)).To(BeTrue(), "Stats should be found")
 		})
+
+		It("runtime should support listing stats for containers filtered by labels [Conformance]", func() {
+			By("create container")
+			labels := map[string]string{"foo": "bar"}
+			containerID := framework.CreateDefaultContainerWithLabels(rc, ic, podID, podConfig, "container-for-stats-with-labels-", labels)
+
+			By("start container")
+			startContainer(rc, containerID)
+
+			By("test container stats")
+			stats := listContainerStats(rc, &runtimeapi.ContainerStatsFilter{LabelSelector: labels})
+			Expect(statFound(stats, containerID)).To(BeTrue(), "Container should be found")
+
+			stats = listContainerStats(rc, &runtimeapi.ContainerStatsFilter{LabelSelector: map[string]string{"foo": "baz"}})
+			Expect(statFound(stats, containerID)).To(BeFalse(), "Container should be filtered")
+		})
 	})
 
 	Context("runtime should support adding volume and device", func() {
@@ -253,14 +266,14 @@ var _ = framework.KubeDescribe("Container", func() {
 
 		AfterEach(func() {
 			By("stop PodSandbox")
-			rc.StopPodSandbox(context.TODO(), podID)
+			Expect(rc.StopPodSandbox(context.TODO(), podID)).NotTo(HaveOccurred())
 			By("delete PodSandbox")
-			rc.RemovePodSandbox(context.TODO(), podID)
+			Expect(rc.RemovePodSandbox(context.TODO(), podID)).NotTo(HaveOccurred())
 		})
 
 		It("runtime should support starting container with volume [Conformance]", func() {
 			By("create host path and flag file")
-			hostPath, _ := createHostPath(podID)
+			hostPath := createHostPath(podID)
 
 			defer os.RemoveAll(hostPath) // clean up the TempDir
 
@@ -272,12 +285,12 @@ var _ = framework.KubeDescribe("Container", func() {
 
 			By("check whether 'hostPath' contains file or dir in container")
 			output := execSyncContainer(rc, containerID, checkPathCmd(hostPath))
-			Expect(len(output)).NotTo(BeZero(), "len(output) should not be zero.")
+			Expect(output).NotTo(BeEmpty(), "len(output) should not be zero.")
 		})
 
 		It("runtime should support starting container with volume when host path is a symlink [Conformance]", func() {
 			By("create host path and flag file")
-			hostPath, _ := createHostPath(podID)
+			hostPath := createHostPath(podID)
 			defer os.RemoveAll(hostPath) // clean up the TempDir
 
 			By("create symlink")
@@ -292,7 +305,7 @@ var _ = framework.KubeDescribe("Container", func() {
 
 			By("check whether 'symlink' contains file or dir in container")
 			output := execSyncContainer(rc, containerID, checkPathCmd(symlinkPath))
-			Expect(len(output)).NotTo(BeZero(), "len(output) should not be zero.")
+			Expect(output).NotTo(BeEmpty(), "len(output) should not be zero.")
 		})
 
 		// TODO(random-liu): Decide whether to add host path not exist test when https://github.com/kubernetes/kubernetes/pull/61460
@@ -309,9 +322,9 @@ var _ = framework.KubeDescribe("Container", func() {
 
 		AfterEach(func() {
 			By("stop PodSandbox")
-			rc.StopPodSandbox(context.TODO(), podID)
+			Expect(rc.StopPodSandbox(context.TODO(), podID)).NotTo(HaveOccurred())
 			By("delete PodSandbox")
-			rc.RemovePodSandbox(context.TODO(), podID)
+			Expect(rc.RemovePodSandbox(context.TODO(), podID)).NotTo(HaveOccurred())
 			By("clean up the TempDir")
 			os.RemoveAll(hostPath)
 		})
@@ -491,15 +504,15 @@ func verifyExecSyncOutput(c internalapi.RuntimeService, containerID string, comm
 }
 
 // createHostPath creates the hostPath and flagFile for volume.
-func createHostPath(podID string) (string, string) {
-	hostPath, err := ioutil.TempDir("", "test"+podID)
+func createHostPath(podID string) (hostPath string) {
+	hostPath, err := os.MkdirTemp("", "test"+podID)
 	framework.ExpectNoError(err, "failed to create TempDir %q: %v", hostPath, err)
 
 	flagFile := "testVolume.file"
 	_, err = os.Create(filepath.Join(hostPath, flagFile))
 	framework.ExpectNoError(err, "failed to create volume file %q: %v", flagFile, err)
 
-	return hostPath, flagFile
+	return hostPath
 }
 
 // createSymlink creates a symlink of path.
@@ -510,7 +523,7 @@ func createSymlink(path string) string {
 }
 
 // createVolumeContainer creates a container with volume and the prefix of containerName and fails if it gets error.
-func createVolumeContainer(rc internalapi.RuntimeService, ic internalapi.ImageManagerService, prefix string, podID string, podConfig *runtimeapi.PodSandboxConfig, hostPath string) string {
+func createVolumeContainer(rc internalapi.RuntimeService, ic internalapi.ImageManagerService, prefix, podID string, podConfig *runtimeapi.PodSandboxConfig, hostPath string) string {
 	By("create a container with volume and name")
 	containerName := prefix + framework.NewUUID()
 	containerConfig := &runtimeapi.ContainerConfig{
@@ -531,10 +544,10 @@ func createVolumeContainer(rc internalapi.RuntimeService, ic internalapi.ImageMa
 }
 
 // createLogContainer creates a container with log and the prefix of containerName.
-func createLogContainer(rc internalapi.RuntimeService, ic internalapi.ImageManagerService, prefix string, podID string, podConfig *runtimeapi.PodSandboxConfig) (string, string) {
+func createLogContainer(rc internalapi.RuntimeService, ic internalapi.ImageManagerService, prefix, podID string, podConfig *runtimeapi.PodSandboxConfig) (logPath, containerID string) {
 	By("create a container with log and name")
 	containerName := prefix + framework.NewUUID()
-	path := fmt.Sprintf("%s.log", containerName)
+	path := containerName + ".log"
 	containerConfig := &runtimeapi.ContainerConfig{
 		Metadata: framework.BuildContainerMetadata(containerName, framework.DefaultAttempt),
 		Image:    &runtimeapi.ImageSpec{Image: framework.TestContext.TestImageList.DefaultTestContainerImage},
@@ -545,10 +558,10 @@ func createLogContainer(rc internalapi.RuntimeService, ic internalapi.ImageManag
 }
 
 // createKeepLoggingContainer creates a container keeps logging defaultLog to output.
-func createKeepLoggingContainer(rc internalapi.RuntimeService, ic internalapi.ImageManagerService, prefix string, podID string, podConfig *runtimeapi.PodSandboxConfig) (string, string) {
+func createKeepLoggingContainer(rc internalapi.RuntimeService, ic internalapi.ImageManagerService, prefix, podID string, podConfig *runtimeapi.PodSandboxConfig) (logPath, containerID string) {
 	By("create a container with log and name")
 	containerName := prefix + framework.NewUUID()
-	path := fmt.Sprintf("%s.log", containerName)
+	path := containerName + ".log"
 	containerConfig := &runtimeapi.ContainerConfig{
 		Metadata: framework.BuildContainerMetadata(containerName, framework.DefaultAttempt),
 		Image:    &runtimeapi.ImageSpec{Image: framework.TestContext.TestImageList.DefaultTestContainerImage},
@@ -558,7 +571,7 @@ func createKeepLoggingContainer(rc internalapi.RuntimeService, ic internalapi.Im
 	return containerConfig.LogPath, framework.CreateContainer(rc, ic, containerConfig, podID, podConfig)
 }
 
-// pathExists check whether 'path' does exist or not
+// pathExists check whether 'path' does exist or not.
 func pathExists(path string) bool {
 	_, err := os.Stat(path)
 	if err == nil {
@@ -571,22 +584,6 @@ func pathExists(path string) bool {
 	return false
 }
 
-// parseDockerJSONLog parses logs in Docker JSON log format.
-// Docker JSON log format example:
-//
-//	{"log":"content 1","stream":"stdout","time":"2016-10-20T18:39:20.57606443Z"}
-//	{"log":"content 2","stream":"stderr","time":"2016-10-20T18:39:20.57606444Z"}
-func parseDockerJSONLog(log []byte, msg *logMessage) {
-	var l jsonlog.JSONLog
-
-	err := json.Unmarshal(log, &l)
-	framework.ExpectNoError(err, "failed with %v to unmarshal log %q", err, l)
-
-	msg.timestamp = l.Created
-	msg.stream = streamType(l.Stream)
-	msg.log = l.Log
-}
-
 // parseCRILog parses logs in CRI log format.
 // CRI log format example :
 //
@@ -621,15 +618,7 @@ func parseLogLine(podConfig *runtimeapi.PodSandboxConfig, logPath string) []logM
 
 	scanner := bufio.NewScanner(f)
 	for scanner.Scan() {
-		line := scanner.Text()
-
-		// to determine whether the log is Docker format or CRI format.
-		if strings.HasPrefix(line, "{") {
-			parseDockerJSONLog([]byte(line), &msg)
-		} else {
-			parseCRILog(line, &msg)
-		}
-
+		parseCRILog(scanner.Text(), &msg)
 		msgLog = append(msgLog, msg)
 	}
 
@@ -642,7 +631,7 @@ func parseLogLine(podConfig *runtimeapi.PodSandboxConfig, logPath string) []logM
 }
 
 // verifyLogContents verifies the contents of container log.
-func verifyLogContents(podConfig *runtimeapi.PodSandboxConfig, logPath string, log string, stream streamType) {
+func verifyLogContents(podConfig *runtimeapi.PodSandboxConfig, logPath, log string, stream streamType) {
 	By("verify log contents")
 	msgs := parseLogLine(podConfig, logPath)
 
@@ -656,6 +645,21 @@ func verifyLogContents(podConfig *runtimeapi.PodSandboxConfig, logPath string, l
 	Expect(found).To(BeTrue(), "expected log %q (stream=%q) not found in logs %+v", log, stream, msgs)
 }
 
+// verifyLogContentsRe verifies the contents of container log using the provided regular expression pattern.
+func verifyLogContentsRe(podConfig *runtimeapi.PodSandboxConfig, logPath, pattern string, stream streamType) {
+	By("verify log contents using regex pattern")
+	msgs := parseLogLine(podConfig, logPath)
+
+	found := false
+	for _, msg := range msgs {
+		if matched, _ := regexp.MatchString(pattern, msg.log); matched && msg.stream == stream {
+			found = true
+			break
+		}
+	}
+	Expect(found).To(BeTrue(), "expected log pattern %q (stream=%q) to match logs %+v", pattern, stream, msgs)
+}
+
 // listContainerStatsForID lists container for containerID.
 func listContainerStatsForID(c internalapi.RuntimeService, containerID string) *runtimeapi.ContainerStats {
 	By("List container stats for containerID: " + containerID)
@@ -664,7 +668,7 @@ func listContainerStatsForID(c internalapi.RuntimeService, containerID string) *
 	return stats
 }
 
-// listContainerStats lists stats for containers based on filter
+// listContainerStats lists stats for containers based on filter.
 func listContainerStats(c internalapi.RuntimeService, filter *runtimeapi.ContainerStatsFilter) []*runtimeapi.ContainerStats {
 	By("List container stats for all containers:")
 	stats, err := c.ListContainerStats(context.TODO(), filter)
