@@ -20,9 +20,8 @@ import (
 	. "github.com/onsi/ginkgo/v2"
 )
 
-// The actual test suite
+// The actual test suite.
 var _ = t.Describe("help", func() {
-
 	const helpMessageIdentifier = "crictl - client for CRI"
 
 	It("should succeed with `help` subcommand", func() {
@@ -42,3 +41,30 @@ var _ = t.Describe("help", func() {
 			"flag provided but not defined")
 	})
 })
+
+// The actual test suite.
+var _ = t.Describe("help subcommand", func() {
+	It("should show help running rm without params", func() {
+		t.CrictlExpectSuccess("rm", "crictl rm command")
+	})
+
+	It("should show help running rmi without params", func() {
+		t.CrictlExpectSuccess("rmi", "crictl rmi command")
+	})
+
+	It("should show help running rmp without params", func() {
+		t.CrictlExpectSuccess("rmp", "crictl rmp command")
+	})
+
+	It("should not show help running rm -a", func() {
+		t.CrictlExpect("rm -a", 0, "", "No containers to remove")
+	})
+
+	It("should not show help running rmi -a", func() {
+		t.CrictlExpect("rmi -a", 0, "", "No images to remove")
+	})
+
+	It("should not show help running rmp -a", func() {
+		t.CrictlExpect("rmp -a", 0, "", "No pods to remove")
+	})
+})
